# -*- coding: utf-8 -*-
import time
import logging
import asyncio
from threading import Lock
from typing import Union, Any, Dict
from aredis import StrictRedis, StrictRedisCluster


class SingletonMeta(type):
    """元类——有限的单例模式
    当初始化参数包含new=True时，将构造一个新的对象
    """
    __instance = None
    __lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls.__lock:
            new = kwargs.pop('new', None)
            if new is True:
                return super().__call__(*args, **kwargs)
            if not cls.__instance:
                cls.__instance = super().__call__(*args, **kwargs)
        return cls.__instance


class MemoryEngine:
    """本地内存作为后端缓存引擎，不支持分布式
    只支持get、set方法
    """

    def __init__(self):
        self.namespace = {}
        self._check_time = 0
        self._check_interval = 60

    def delete(self, key: str) -> None:
        """删除指定缓存"""
        if key in self.namespace:
            del self.namespace[key]

    def et_clear(self) -> None:
        """清理超时缓存"""
        clear_names = []
        if time.time() > self._check_time + self._check_interval:
            self._check_time = time.time()
            for name, block in self.namespace.items():
                if block.ttl < -1:
                    clear_names.append(name)
        for name in clear_names:
            del self.namespace[name]

    async def ttl(self, name) -> int:
        self.et_clear()
        if name not in self.namespace:
            return -1
        return int(self.namespace[name].ttl)

    async def get(self, name):
        self.et_clear()
        if name not in self.namespace:
            return None
        return self.namespace[name].val

    async def set(self, name, value, ex=None, px=None, nx=False, xx=False):
        if nx and name in self.namespace:
            return
        if xx and name not in self.namespace:
            return
        self.namespace[name] = DataBlock(name, value, ex, px)
        if len(value) > 16384 and (ex or px):
            # 实验性功能 大容量缓存清理机制 避免长时间不使用缓存下占用内存
            life = ex if ex else px // 1000
            loop = asyncio.get_event_loop()
            loop.call_later(life * 2, self.delete, name)


class DataBlock:
    """内存数据块 封装了有效期"""

    def __init__(self, name: str, value: Any, ex: float = None,
                 px: float = None):
        """
        :param name: key名
        :param value: 存储value
        :param ex: 生命周期，单位秒
        :param px: 生命周期，单位毫秒
        """
        self._name = name
        self._value = value
        self.et = time.time() - 1
        if ex:
            self.et += ex
        if px:
            self.et += (px / 1000)
        if not ex and not px:
            self._ttl = -1

    @property
    def val(self):
        return self._value if self.ttl >= -1 else None

    @property
    def ttl(self):
        if hasattr(self, '_ttl'):
            return self._ttl
        return self.et - time.time()

    def __repr__(self):
        return f'<name={self._name}>'


class Cache(metaclass=SingletonMeta):
    """一个基于redis封装的异步缓存类，它可以快速方便切换多个缓存库
    Cache类默认使用default缓存库，你可以使用select(db_name)切换其他库，并且select支持
    链式调用，但select方法并不会改变原对象指向的default缓存库
    Cache对象通过反射拥有了StrictRedis和StrictRedisCluster类下的所有方法，你可以直接对
    对象执行redis命令，此外Cache还封装了一个方法execute(command, *args, **kwargs)
    相比于反射方法，使用execute方法会自动对返回数据解码
    针对字符串类型，Cache对get和set方法作了优化，当使用get和set方法时，可以同时传递一个序列化器，
    它会查询和存储时自动使用序列化器，也就是说你可以使用set方法存储任意序列化器支持的对象
    """
    logger = logging.getLogger(__name__)

    def __init__(self, config: dict):
        """
        :param config: 缓存数据库字典
        :return: Cache对象
        """
        self._default = 'default'
        self._caches = {}
        serializer = config.pop('serializer', 'ujson')
        try:
            self.serializer = __import__(serializer)
        except:
            self.serializer = __import__('json')
        for key, value in config.items():
            try:
                if value.get('engine') == 'memory':
                    self._caches[key] = MemoryEngine()
                elif 'startup_nodes' in value:
                    self._caches[key] = StrictRedisCluster(**value)
                else:
                    self._caches[key] = StrictRedis(**value)
            except Exception as e:
                self.logger.error(e)

    @property
    def all(self) -> Dict[str, Union[StrictRedis, StrictRedisCluster]]:
        """返回全部缓存数据库"""
        return self._caches

    @property
    def current_db(self) -> Union[StrictRedis, StrictRedisCluster]:
        """返回缓存对象指向的缓存数据库"""
        return self._caches[self._default]

    def select(self, name: str = 'default') -> 'Cache':
        """获取指定缓存数据库
        支持多次链式调用select方法
        永远不会改变app所绑定的默认缓存数据库
        :param name: 定义的数据库名，默认值为"default"
        :return: Cache对象
        """
        if name not in self._caches:
            raise AttributeError(f'Cache database "{name}" not found. '
                                 f'Please check CACHES config in settings')
        obj = Cache(config={}, new=True)
        obj._caches = self._caches
        obj._default = name
        return obj

    async def execute(self, command: str, *args, **kwargs) -> Any:
        """实现结果自解码
        :param command: 执行的redis原生命令
        :return: 返回redis结果的utf8解码
        """
        if hasattr(StrictRedis, command):
            result = await getattr(self, command)(*args, **kwargs)
            if result:
                result = result.decode('utf8')
            return result
        else:
            raise getattr(self, command)

    async def get(self, name, serializer=None, **kwargs) -> Any:
        """覆盖redis的字符串get方法，提供序列化能力
        :param name: key
        :param serializer: 使用指定的序列化模块
        :param kwargs: 传递给序列化方法
        :return: 返回redis结果的反序列化对象
        """
        if not serializer:
            serializer = self.serializer
        value = await self.current_db.get(name)
        if not value:
            return None
        else:
            if isinstance(value, bytes):
                value = value.decode('utf8')
        try:
            return serializer.loads(value, **kwargs)
        except ValueError:
            return value

    async def set(self, name: str, value: Any, serializer=None,
                  ex=None, px=None, nx=False, xx=False, **kwargs) -> bool:
        """永远在redis层以string格式存储，提供反序列化能力
        :param name:
        :param value:
        :param serializer: 使用指定的序列化模块
        :param ex: 设置键key的过期时间，单位为秒
        :param px: 设置键key的过期时间，单位为毫秒
        :param nx: 只有键key不存在的时候才会设置key的值
        :param xx: 只有键key存在的时候才会设置key的值
        :param kwargs: 传递给反序列化方法
        :return: 执行结果
        """
        if not serializer:
            serializer = self.serializer
        _kwargs = {'ensure_ascii': True}
        _kwargs.update(kwargs)
        if not isinstance(value, str):
            value = serializer.dumps(value, **_kwargs)
        return await self.current_db.set(name, value, ex, px, nx, xx)

    def handle(self, backed: str):
        # TODO 查询缓存库

        # TODO 执行handler获取结果

        # TODO 返回结果并存储至缓存
        """提供给视图方法的装饰器 它缓存视图方法返回的结果"""

    def __getitem__(self, item) -> 'Cache':
        return self.select(item)

    def __getattr__(self, attr) -> Any:
        return getattr(self.current_db, attr)



# -*- coding: utf-8 -*-
import io
import time

from openpyxl import Workbook, load_workbook
from openpyxl.writer.excel import save_virtual_workbook
from sanic.response import raw

from exceptions import APIException
from mixins import ListModelMixin, CreateModelMixin
from status import HttpStatus
from viewsets import GenericViewSet


class ExportImportView(ListModelMixin, CreateModelMixin, GenericViewSet):
    """导入导出基类"""
    export_header_title = {}
    import_header_title = {}
    upload_file_key = 'import_file'

    async def list(self, request, *args, **kwargs):
        queryset = await self.get_queryset()
        page = await self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
        else:
            serializer = self.get_serializer(queryset, many=True)

        virtual_workbook = self.write_data_to_excel(await serializer.data)

        return raw(save_virtual_workbook(virtual_workbook), headers={
            'Content-Disposition': 'attachment;filename={}'.format(self.get_export_file_name())
        })

    async def create(self, request, *args, **kwargs):
        workbook: Workbook = self.load_workbook_by_request()
        worksheet = workbook.worksheets[0]
        rows = worksheet.max_row  # 获取表的最大行数
        columns = worksheet.max_column  # 获取表的最大列数

        # 检查导入数据是否必备
        for column in range(1, columns + 1):
            chinese_key = worksheet.cell(1, column).value
            if chinese_key not in self.import_header_title:
                return self.error_json_response('存在意外的数据列%s' % chinese_key)

        for row in range(2, rows + 1):
            row_record = {}
            for column in range(1, columns + 1):
                chinese_key = worksheet.cell(1, column).value
                letter_key = self.import_header_title[chinese_key]
                row_record[letter_key] = worksheet.cell(row, column).value
            serializer = self.serializer_class(data=row_record)
            await serializer.is_valid(raise_exception=True)
            await self.perform_create(serializer)
        return self.success_json_response(msg='导入成功', http_status=HttpStatus.HTTP_201_CREATED)

    def load_workbook_by_request(self):
        """从请求中加载Excel"""
        upload_file = self.request.files.get(self.upload_file_key)
        if upload_file:
            raise APIException(message='必须上传导入文件 file_name.xlsx', http_status=HttpStatus.HTTP_200_OK)
        stream = upload_file.body
        workbook: Workbook = load_workbook(io.BytesIO(stream))
        return workbook

    def get_export_file_name(self):
        """导出名"""
        return 'export_{}.xlsx'.format(int(time.time()))

    def create_excel(self):
        """导出执行"""
        virtual_workbook = Workbook()
        virtual_workbook.create_sheet('export', 0)
        virtual_workbook.remove_sheet(virtual_workbook.worksheets[1])
        return virtual_workbook

    def write_data_to_excel(self, data):
        """为excel写入数据"""
        virtual_workbook = self.create_excel()
        sheet = virtual_workbook.worksheets[0]
        for x_index, key in enumerate(self.export_header_title):
            x = x_index + 1
            title = self.export_header_title[key]
            sheet[self.get_index(x, 1)] = title
            for y_index, row in enumerate(data):
                y = y_index + 2
                sheet[self.get_index(x, y)] = row[key]
        return virtual_workbook

    def get_index(self, x, y):
        """得到字母+数字坐标"""
        return '{}{}'.format(self.get_letter_head(x), y)

    def get_letter_head(self, x):
        """数字转字母"""
        start_num = ord('A')
        if x <= 26:
            return chr(start_num + x - 1)
        else:
            return self.get_letter_head((x - 1) // 26) + chr(start_num + (x - 1) % 26)



# -*- coding: utf-8 -*-

from srf.views import APIView
from srf.filters import ORMAndFilter
from srf.exceptions import APIException
from srf.status import HttpStatus
from srf import mixins

__all__ = ('GenericAPIView', 'CreateAPIView', 'ListAPIView', 'RetrieveAPIView',
           'DestroyAPIView', 'UpdateAPIView', 'ListCreateAPIView', 'RetrieveUpdateAPIView',
           'RetrieveDestroyAPIView', 'RetrieveUpdateDestroyAPIView')


class GenericAPIView(APIView):
    """
    视图集视图，可以配合Mixin实现复杂的视图集，
    数据来源基于模型查询集,可以配合Route组件实现便捷的路由管理
    """
    # 数据来源
    queryset = None
    # 数据序列化器
    serializer_class = None
    # 主键
    lookup_field = 'pk'

    # 分页器
    pagination_class = None

    # 过滤器
    filter_class = ORMAndFilter
    search_fields = None

    async def get_object(self):
        """
        返回视图显示的对象。
        如果您需要提供非标准的内容，则可能要覆盖此设置
        queryset查找。
        """
        queryset = await self.get_queryset()

        lookup_field = self.lookup_field

        assert lookup_field in self.kwargs, (
                '%s 不存在于 %s 的 Url配置中的关键词内 ' %
                (lookup_field, self.__class__.__name__,)
        )

        filter_kwargs = {lookup_field: self.kwargs[lookup_field]}
        obj = await queryset.get_or_none(**filter_kwargs)
        if obj is None:
            raise APIException('不存在%s为%s的数据' % (lookup_field, self.kwargs[lookup_field]),
                               http_status=HttpStatus.HTTP_200_OK)

        # May raise a permission denied
        await self.check_object_permissions(self.request, obj)

        return obj

    async def get_queryset(self):
        assert self.queryset is not None, (
                "'%s'应该包含一个'queryset'属性，"
                "或重写`get_queryset()`方法。"
                % self.__class__.__name__
        )
        queryset = self.queryset
        # filter_orm = await self.filter_orm()
        # queryset = queryset.filter(filter_orm)
        return queryset

    async def get_filter_queryset(self):
        filter_orm = await self.filter_orm()
        queryset = await self.get_queryset()
        filter_queryset = queryset.filter(filter_orm)
        return filter_queryset

    async def filter_orm(self):
        """得到ORM过滤参数"""
        return self.filter_class(self.request, self).orm_filter

    def get_serializer(self, *args, **kwargs):
        """
        返回应该用于验证和验证的序列化程序实例
        对输入进行反序列化，并对输出进行序列化。
        """
        serializer_class = self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        setattr(serializer_class, 'context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self):
        """
        返回用于序列化器的类。
        默认使用`self.serializer_class`。

        如果您需要提供其他信息，则可能要覆盖此设置
        序列化取决于传入的请求。

        （例如，管理员获得完整的序列化，其他获得基本的序列化）
        """
        assert self.serializer_class is not None, (
                "'%s' should either include a `serializer_class` attribute, "
                "or override the `get_serializer_class()` method."
                % self.__class__.__name__
        )
        return self.serializer_class

    def get_serializer_context(self):
        """
        提供给序列化程序类的额外上下文。
        """
        return {
            'request': self.request,
            'view': self
        }

    @property
    def paginator(self):
        """
        与视图关联的分页器实例，或“None”。
        """
        if not hasattr(self, '_paginator'):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class(self.request, self)
        return self._paginator

    async def get_paginator_count(self, queryset):
        """
        获取记录总数
        :param queryset:
        :return:
        """
        return await queryset.count()

    async def paginate_queryset(self, queryset):
        """
        返回单页结果，如果禁用了分页，则返回“无”。
        """
        if self.paginator is None:
            return None
        offset = (self.paginator.query_page - 1) * self.paginator.query_page_size
        return queryset.limit(self.paginator.query_page_size).offset(offset)

    def get_paginated_response(self, data):
        """
        返回给定输出数据的分页样式`Response`对象。
        """
        return {
            'total': self.paginator.count,
            'pageNo': self.paginator.query_page,
            'pageSize': self.paginator.query_page_size,
            'list': data
        }


class CreateAPIView(mixins.CreateModelMixin,
                    GenericAPIView):
    """
    用于创建模型实例的具体视图。
    """

    async def post(self, request, *args, **kwargs):
        return await self.create(request, *args, **kwargs)


class ListAPIView(mixins.ListModelMixin,
                  GenericAPIView):
    """
    列出查询集的具体视图。
    """

    async def get(self, request, *args, **kwargs):
        return await self.list(request, *args, **kwargs)


class RetrieveAPIView(mixins.RetrieveModelMixin,
                      GenericAPIView):
    """
    用于检索模型实例的具体视图。
    """

    async def get(self, request, *args, **kwargs):
        return await self.retrieve(request, *args, **kwargs)


class DestroyAPIView(mixins.DestroyModelMixin,
                     GenericAPIView):
    """
    用于删除模型实例的具体视图。
    """

    async def delete(self, request, *args, **kwargs):
        return await self.destroy(request, *args, **kwargs)


class UpdateAPIView(mixins.UpdateModelMixin,
                    GenericAPIView):
    """
    用于更新模型实例的具体视图。
    """

    async def put(self, request, *args, **kwargs):
        return await self.update(request, *args, **kwargs)

    async def patch(self, request, *args, **kwargs):
        return await self.partial_update(request, *args, **kwargs)


class ListCreateAPIView(mixins.ListModelMixin,
                        mixins.CreateModelMixin,
                        GenericAPIView):
    """
    用于列出查询集或创建模型实例的具体视图。
    """

    async def get(self, request, *args, **kwargs):
        return await self.list(request, *args, **kwargs)

    async def post(self, request, *args, **kwargs):
        return await self.create(request, *args, **kwargs)


class RetrieveUpdateAPIView(mixins.RetrieveModelMixin,
                            mixins.UpdateModelMixin,
                            GenericAPIView):
    """
    用于检索、更新模型实例的具体视图。
    """

    async def get(self, request, *args, **kwargs):
        return await self.retrieve(request, *args, **kwargs)

    async def put(self, request, *args, **kwargs):
        return await self.update(request, *args, **kwargs)

    async def patch(self, request, *args, **kwargs):
        return await self.partial_update(request, *args, **kwargs)


class RetrieveDestroyAPIView(mixins.RetrieveModelMixin,
                             mixins.DestroyModelMixin,
                             GenericAPIView):
    """
    用于检索或删除模型实例的具体视图。
    """

    async def get(self, request, *args, **kwargs):
        return await self.retrieve(request, *args, **kwargs)

    async def delete(self, request, *args, **kwargs):
        return await self.destroy(request, *args, **kwargs)


class RetrieveUpdateDestroyAPIView(mixins.RetrieveModelMixin,
                                   mixins.UpdateModelMixin,
                                   mixins.DestroyModelMixin,
                                   GenericAPIView):
    """
    用于检索、更新或删除模型实例的具体视图。
    """

    async def get(self, request, *args, **kwargs):
        return await self.retrieve(request, *args, **kwargs)

    async def put(self, request, *args, **kwargs):
        return await self.update(request, *args, **kwargs)

    async def patch(self, request, *args, **kwargs):
        return await self.partial_update(request, *args, **kwargs)

    async def delete(self, request, *args, **kwargs):
        return await self.destroy(request, *args, **kwargs)



# -*- coding: utf-8 -*-
from srf.status import HttpStatus, RuleStatus


class CacheExpireEXC(Exception):
    def __init__(self, message="缓存超时", *args, **kwargs):
        self.message = message

    def __str__(self) -> str:
        return '<CacheExpireEXC: {}>'.format(self.message)


class CacheNoFoundEXC(Exception):
    def __init__(self, message="找不到指定引擎", *args, **kwargs):
        self.message = message

    def __str__(self) -> str:
        return '<CacheNoFoundEXC: {}>'.format(self.message)


class ValidatorAssertError(Exception):
    pass


class ValidationException(Exception):
    """验证错误类"""
    default_detail = '无效的输入'
    default_code = 'invalid'

    def __init__(self, error_detail=None, code=None):
        if error_detail is None:
            error_detail = self.default_detail
        if code is None:
            code = self.default_code

        if not isinstance(error_detail, dict) and not isinstance(error_detail, list):
            error_detail = [error_detail]

        self.code = code
        self.error_detail = error_detail


class APIException(Exception):
    def __init__(self, message, status=RuleStatus.STATUS_0_FAIL, http_status=HttpStatus.HTTP_500_INTERNAL_SERVER_ERROR,
                 *args, **kwargs):
        self.message = message
        self.status = status
        self.http_status = http_status

    def response_data(self):
        return {
            'msg': self.message,
            'status': self.status,
            'http_status': self.http_status
        }


class PermissionDenied(APIException):
    def __init__(self, message='没有操作权限，请完成对应授权', status=RuleStatus.STATUS_0_FAIL, http_status=HttpStatus.HTTP_403_FORBIDDEN,
                 *args, **kwargs):
        super().__init__(message, status, http_status, *args, **kwargs)


class AuthenticationDenied(APIException):
    def __init__(self, message='身份验证失败，请重新登录', status=RuleStatus.STATUS_0_FAIL, http_status=HttpStatus.HTTP_401_UNAUTHORIZED,
                 *args, **kwargs):
        super().__init__(message, status, http_status, *args, **kwargs)


# -*- coding: utf-8 -*-
from srf.paginations import ORMPagination
from srf.status import HttpStatus

__all__ = (
    'ListModelMixin', 'CreateModelMixin', 'RetrieveModelMixin', 'UpdateModelMixin', 'DestroyModelMixin'
)


class ListModelMixin:
    """
    适用于输出列表类型数据
    """
    pagination_class = ORMPagination
    detail = False

    async def get(self, request, *args, **kwargs):
        return await self.list(request, *args, **kwargs)

    async def list(self, request, *args, **kwargs):
        # queryset = self.get_queryset()
        queryset = await self.get_filter_queryset()

        no_page = request.args.get("no_page", False)
        if no_page in ('true', True):
            self.pagination_class = None
        page = await self.paginate_queryset(queryset)
        if page is not None:
            self.paginator.set_count(await self.get_paginator_count(queryset))
            serializer = self.get_serializer(page, many=True)
            data = self.get_paginated_response(await serializer.data)
            return self.success_json_response(msg="请求成功！", data=data)

        serializer = self.get_serializer(queryset, many=True)
        return self.success_json_response(msg="请求成功！", data=await serializer.data)


class CreateModelMixin:
    """
    适用于快速创建内容
    占用 post 方法
    """

    async def post(self, request, *args, **kwargs):
        return await self.create(request, *args, **kwargs)

    async def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        await serializer.is_valid(raise_exception=True)
        await self.perform_create(serializer)
        return self.success_json_response(msg="创建成功！", data=await serializer.data,
                                          http_status=HttpStatus.HTTP_201_CREATED)

    async def perform_create(self, serializer):
        await serializer.save()


class RetrieveModelMixin:
    """
    适用于查询指定PK的内容
    """
    detail = True

    async def get(self, request, *args, **kwargs):
        return await self.retrieve(request, *args, **kwargs)

    async def retrieve(self, request, *args, **kwargs):
        instance = await self.get_object()
        serializer = self.get_serializer(instance)
        return self.success_json_response(msg="查询成功！", data=await serializer.data)


class UpdateModelMixin:
    """
    适用于快速创建更新操作
    """

    async def put(self, request, *args, **kwargs):
        return await self.update(request, *args, **kwargs)

    async def patch(self, request, *args, **kwargs):
        return await self.partial_update(request, *args, **kwargs)

    async def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = await self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        await serializer.is_valid(raise_exception=True)
        await self.perform_update(serializer)
        return self.success_json_response(msg="更新成功！", data=await serializer.data)

    async def perform_update(self, serializer):
        await serializer.save()

    async def partial_update(self, request, *args, **kwargs):
        kwargs['partial'] = True
        return await self.update(request, *args, **kwargs)


class DestroyModelMixin:
    """
    用于快速删除
    """

    async def delete(self, request, *args, **kwargs):
        return await self.destroy(request, *args, **kwargs)

    async def destroy(self, request, *args, **kwargs):
        instance = await self.get_object()
        await self.perform_destroy(instance)
        return self.success_json_response(msg="删除成功！", http_status=HttpStatus.HTTP_200_OK)

    async def perform_destroy(self, instance):
        await instance.delete()



from srf.exceptions import PermissionDenied
from srf.constant import ALL_METHOD


class BasePermission:
    async def has_permission(self, request, view):
        pass

    async def has_obj_permission(self, request, view, obj):
        pass


class ViewMapPermission(BasePermission):
    """
    APIView permission_classes

    :example
        class ViewName(APIView):
            permission_classes = (ViewMapPermission)
            permission_map = {
                'get':['can_view']
                'post':['can_create']
                ....
                'all':['need_perm']
            }

    Note, request.user must need `has_permissions(codes)` method

    :param BasePermission: BasePermission
    """

    def get_permission_map(self, view):
        permission_map = self.permission_map
        if hasattr(view, 'permission_map'):
            permission_map.update(view.permission_map)
        return permission_map

    async def has_permission(self, request, view):
        permission_map = self.get_permission_map(view)
        all_permissions = permission_map.get('all')
        method = request.method.lower()
        method_permission = permission_map.get(method)
        permissions = (*method_permission, *all_permissions)
        if not request.user.has_permissions(permissions):
            raise PermissionDenied()

    async def has_obj_permission(self, request, view, obj):
        pass

    @property
    def permission_map(self):
        permission_map = {method.lower(): () for method in ALL_METHOD}
        permission_map['all'] = ()
        return permission_map



# -*- coding: utf-8 -*-
from collections import namedtuple

# 默认分组

Route = namedtuple('Route', ['url', 'mapping', 'name', 'detail', 'initkwargs', 'is_base'])


class BaseRoute:

    def __init__(self, trailing_slash=False):
        """
        基础类
        :param trailing_slash:
        """
        self.trailing_slash = '/' if trailing_slash else ''
        self.registry = []

    def register(self, viewset: object, prefix: str, basename: str = None, is_base: bool = False):
        if basename is None:
            basename = prefix.replace('/', '_')
        self.registry.append((prefix, viewset, basename, is_base))

        # invalidate the urls cache
        if hasattr(self, '_urls'):
            del self._urls

    def get_urls(self):
        pass

    @property
    def urls(self):
        if not hasattr(self, '_urls'):
            self._urls = self.get_urls()
        return self._urls


class ViewSetRouter(BaseRoute):
    routes = [
        # List route.
        Route(
            url=r'{prefix}{trailing_slash}',
            mapping={
                'get': 'list',
                'post': 'create'
            },
            name='{basename}-list',
            detail=False,
            initkwargs={'suffix': 'List'},
            is_base=False,
        ),
        Route(
            url=r'{prefix}/<{lookup}:string>{trailing_slash}',
            mapping={
                'get': 'retrieve',
                'put': 'put',
                'patch': 'patch',
                'delete': 'destroy'
            },
            name='{basename}-detail',
            detail=True,
            initkwargs={'suffix': 'Instance'},
            is_base=False,
        ),
    ]

    def get_lookup(self, viewset):
        """
        得到主键字段名
        :param viewset: 集合视图
        :return:
        """
        return getattr(viewset, 'lookup_field', 'pk')

    def get_urls(self):
        """
        Use the registered viewsets to generate a list of URL patterns.
        """
        ret = []

        for prefix, viewset, basename, is_base in self.registry:
            lookup = self.get_lookup(viewset)

            for route in self.routes:

                # Only actions which actually exist on the viewset will be bound
                mapping = self.get_method_map(viewset, route.mapping)
                if not mapping:
                    continue

                # Build the url pattern
                uri = route.url.format(
                    prefix=prefix,
                    lookup=lookup,
                    trailing_slash=self.trailing_slash
                )

                # If there is no prefix, the first part of the url is probably
                #   controlled by project's urls.py and the router is in an app,
                #   so a slash in the beginning will (A) cause Django to give
                #   warnings and (B) generate URLS that will require using '//'.
                if not prefix and uri[:2] == '^/':
                    uri = '^' + uri[2:]

                initkwargs = route.initkwargs.copy()
                initkwargs.update({
                    'basename': basename,
                    'detail': route.detail,
                })

                view = viewset.as_view(mapping, **initkwargs)
                name = route.name.format(basename=basename)
                ret.append({
                    'handler': view,
                    'uri': uri,
                    'name': name,
                    'is_base': is_base,

                })

        return ret

    def get_method_map(self, viewset, method_map):
        """得到可用的模型"""
        bound_methods = {}
        for method, action in method_map.items():
            if hasattr(viewset, action):
                bound_methods[method] = action
        return bound_methods


from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()
setup(
    name="srf_hdkj",
    version="0.3",
    author="hdkj",
    author_email="",
    description="SDK about postgres",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=find_packages(),
    install_requires=['ujson==5.1.0', 'openpyxl==3.0.9'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
